#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utilities for Slothstock.

File: utils.py
Author: huxuan
Email: i(at)huxuan.org
"""

from wxpusher import WxPusher


def export_ebk(symbols, filename='sloth.ebk'):
    """Export symbols in ebk format."""
    res = [b'']
    for symbol in symbols:
        code, exchange = symbol.split('.')
        if exchange == 'SH':
            res.append(f'1{code}'.encode())
        else:
            res.append(f'0{code}'.encode())

    with open(filename, 'wb') as fout:
        fout.write(b'\r\n'.join(res))


def import_ebk(filename):
    """Import symbols from ebk format."""
    res = []
    with open(filename, 'rb') as fin:
        fin.readline()  # The first line is useless.
        for line in fin:
            line = line.decode().strip()
            if line[0] == '1':
                res.append(f'{line[1:]}.SH')
            else:
                res.append(f'{line[1:]}.SZ')
    return res


def send_notification(content, token, topic_ids, uids):
    """Send notification."""
    if not token:
        return
    if not uids and not topic_ids:
        return
    WxPusher.send_message(content, uids=uids, topic_ids=topic_ids, token=token)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Xueqiu provider.

File: xueqiu.py
Author: huxuan
Email: i(at)huxuan.org
"""
from datetime import datetime
from datetime import timedelta

from faker import Faker
import pandas
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .. import constants
from .. import exceptions

XUEQIU_INDEX_URL = 'https://xueqiu.com/'
XUEQIU_KLINE_URL = 'https://stock.xueqiu.com/v5/stock/chart/kline.json'
XUEQIU_LIST_URL = 'https://xueqiu.com/service/v5/stock/screener/quote/list'
XUEQIU_ETF_LIST_URL = 'https://xueqiu.com/service/v5/stock/screener/fund/list'


def create_fake_session(index_url=None):
    """Create fake seesion."""
    session = requests.Session()
    session.headers[constants.HEADER_USER_AGENT] = Faker().internet_explorer()
    if index_url:  # Browser index page for cookies.
        requests_retry_session(session).get(index_url)
    return session


def requests_retry_session(
        session=None,
        retries=3,
        backoff_factor=0.3,
        status_forcelist=(500, 502, 504)):
    """Send requests with retry."""
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def symbol_transform(symbol):
    """Transform symbol for XueQiu."""
    return ''.join(symbol.split('.')[::-1])


def symbol_restore(symbol):
    """Restore symbol for XueQiu."""
    return f'{symbol[2:]}.{symbol[:2]}'


def datetime_to_timestamp(dt_origin=None):
    """Convert local datetime to utc timestamp."""
    if dt_origin is None:
        dt_origin = datetime.utcnow() + timedelta(hours=8)
    elif isinstance(dt_origin, str):
        dt_origin = datetime.strptime(dt_origin, constants.DATETIME_FORMAT)
    elif not isinstance(dt_origin, datetime):
        raise exceptions.InvalidDatetimeError(dt_origin)
    return int(dt_origin.timestamp() * 1000)


def timestamp_to_datetime(timestamp):
    """Convert utc timestamp to local datetime."""
    return datetime.fromtimestamp(timestamp / 1000)


def check_response(res):
    """Roughly check result in common."""
    if not res:
        raise exceptions.NoResponseError(res.status_code)

    try:
        res = res.json()
    except ValueError:
        return exceptions.InvalidResultError(res.content)

    if res.get('error_code') != 0:
        return exceptions.XueQiuError(
            res['error_code'],
            res['error_description'])

    if not res.get('data'):
        raise exceptions.NoResultError(res.content)

    return res


class MetaXueQiu(type):
    """Metaclass for XueQiu."""

    session = create_fake_session(XUEQIU_INDEX_URL)

    @classmethod
    def kline(cls, symbol, period='day', begin=None, count=142):
        """Fetch kline information."""
        params = {
            'begin': datetime_to_timestamp(begin),
            'count': -abs(count),
            'indicator': 'kline',
            'period': period,
            'symbol': symbol_transform(symbol),
            'type': 'before',
        }
        res = requests_retry_session(cls.session).get(
            XUEQIU_KLINE_URL, params=params)
        res = check_response(res)

        data = pandas.DataFrame(
            data=res['data'].get('item'),
            columns=res['data'].get('column'))
        data.rename(columns={'timestamp': 'datetime'}, inplace=True)
        data.datetime = data.datetime.apply(timestamp_to_datetime)
        data.set_index('datetime', inplace=True)
        return data

    @classmethod
    def list(cls, url=XUEQIU_LIST_URL, **params):
        """List stock information."""
        params.update({
            '_': datetime_to_timestamp(),
            'order': params.get('reverse', False) and 'desc' or 'asc',
            'order_by': params.get('order_by', 'symbol'),
            'page': params.get('page', 1),
            'size': params.get('size', 4000),
            'type': params.get('type', 'sh_sz'),
        })
        res = requests_retry_session(cls.session).get(url, params=params)
        res = check_response(res)

        data = pandas.DataFrame(res['data']['list'])
        data.symbol = data.symbol.apply(symbol_restore)
        data.set_index('symbol', inplace=True)
        return data

    @classmethod
    def list_ashare(cls, **params):
        """List A Share stocks."""
        data = cls.list(**params)
        return data[data.index.str[:2].isin(constants.STOCK_PREFIX_MARKET)]

    @classmethod
    def list_etf(cls, **params):
        """List ETF."""
        params.update({
            'parent_type': 1,
            'size': params.get('size', 300),
            'type': 18,
        })
        return cls.list(XUEQIU_ETF_LIST_URL, **params)


class XueQiu(metaclass=MetaXueQiu):  # pylint: disable=too-few-public-methods
    """XueQiu provider."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unittests for utils.

File: test_utils.py
Author: huxuan
Email: i(at)huxuan.org
"""
import random
import tempfile
import unittest

from slothstock import utils
from slothstock.providers.xueqiu import XueQiu


class TestUtils(unittest.TestCase):
    """Unittests for utils."""

    def test_import_export_ebk(self):
        """Positive case for import & export ebk."""
        temp_file = tempfile.NamedTemporaryFile()
        stocks = XueQiu.list_ashare(
            page=random.randint(1, 20),
            size=random.randint(1, 100))
        utils.export_ebk(stocks.index, temp_file.name)
        symbols = utils.import_ebk(temp_file.name)
        for symbol in symbols:
            self.assertIn(symbol, stocks.index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom constants.

File: constants.py
Author: huxuan
Email: i(at)huxuan.org
"""

DATETIME_FORMAT = '%Y-%m-%d %H:%M'
FORMAT_EBK = 'ebk'
HEADER_USER_AGENT = 'User-Agent'
PERIODS = [
    '1m',
    '5m',
    '15m',
    '60m',
    'day',
    'week',
    'month',
    'quarter',
    'year',
]
PERIODS_VALID = PERIODS[1:5]
STOCK_PREFIX_MARKET = {
    '00': 'SZ',
    '15': 'SZ',
    '30': 'SZ',
    '51': 'SH',
    '60': 'SH',
    '68': 'SH',
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unittests for MACD.

File: test_macd_indicator.py
Author: huxuan
Email: i(at)huxuan.org
"""
from datetime import datetime
import unittest

from slothstock.indicators import macd_indicator
from slothstock.providers.xueqiu import XueQiu


class TestMACD(unittest.TestCase):
    """Unittests for MACD."""

    def test_is_about_to_top_divergence(self):
        """Positive case for top divergence."""
        res = XueQiu.kline('000016.SH', 'day', datetime(2019, 10, 21))
        macd, macdsignal, macdhist = macd_indicator.clean_macd(res.close)
        self.assertTrue(macd_indicator.is_positive(macd, macdsignal))
        self.assertTrue(macd_indicator.is_about_to_top_divergence(
            macd, macdsignal, macdhist))

    def test_is_about_to_bottom_divergence(self):
        """Positive case for bottom divergence."""
        res = XueQiu.kline('000016.SH', 'week', datetime(2019, 1, 11))
        macd, macdsignal, macdhist = macd_indicator.clean_macd(res.close)
        self.assertTrue(macd_indicator.is_negative(macd, macdsignal))
        self.assertTrue(macd_indicator.is_about_to_bottom_divergence(
            macd, macdsignal, macdhist))

    def test_is_expand_golden_cross(self):
        """Positive case for expand golden cross."""
        res = XueQiu.kline('000001.SH', 'day', '2020-01-03 00:00')
        _, _, macdhist = macd_indicator.clean_macd(res.close)
        self.assertTrue(macd_indicator.is_expand_golden_cross(macdhist))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Script indicating buy signals.

File: sloth_buy.py
Author: huxuan
Email: i(at)huxuan.org
"""
import argparse
from datetime import datetime
from datetime import timedelta
import time

import numpy
import pandas
from tqdm import tqdm

from slothstock import __version__
from slothstock import constants
from slothstock import exceptions
from slothstock import utils
from slothstock.indicators import macd_indicator
from slothstock.providers.xueqiu import XueQiu


def parse_args():
    """Argument Parser."""
    parser = argparse.ArgumentParser()
    parser.add_argument('--token')
    parser.add_argument('--reserve-st', action='store_false')
    parser.add_argument('--reserve-suspend', action='store_false')
    parser.add_argument('--topic-ids', action='append', default=[])
    parser.add_argument('--uids', action='append', default=[])
    parser.add_argument('-d', '--daemon', action='store_true')
    parser.add_argument('-e', '--ebk', action='append', default=[])
    parser.add_argument('-i', '--interval', default=0.1, type=float)
    parser.add_argument('-o', '--output', default='sloth.ebk')
    parser.add_argument('-p', '--period', default='day')
    parser.add_argument('-s', '--strict', action='store_true')
    parser.add_argument('-t', '--title')
    parser.add_argument('-v', '--version', action='version',
                        version=__version__)
    return parser.parse_args()


def load_stocks(ebk):
    """Load symbols."""
    stocks_ashare = XueQiu.list_ashare()
    stocks_etf = XueQiu.list_etf()
    stocks = pandas.concat([stocks_ashare, stocks_etf], sort=False)

    if ebk:
        res = set()
        for filename in ebk:
            res.update(utils.import_ebk(filename))
        stocks = stocks[stocks.index.isin(res)]

    return stocks


def check_buy(stocks, args):
    """Check buy signals."""
    res = set()
    idx = constants.PERIODS.index(args.period)
    symbols = stocks.sample(frac=1).index

    pbar = tqdm(symbols, disable=args.daemon)
    for symbol in pbar:
        pbar.set_description(symbol)

        # Whether to reserve ST stocks.
        if args.reserve_st and 'name' in stocks.columns and \
                'ST' in stocks.loc[symbol, 'name']:
            continue

        # Whether to reserve suspended stocks.
        if args.reserve_suspend and 'volume' in stocks.columns and \
                numpy.isnan(stocks.loc[symbol, 'volume']):
            continue

        # Check great-great-grandparent period.
        if args.strict:
            time.sleep(args.interval)
            period_cur = constants.PERIODS[idx + 4]
            kline = XueQiu.kline(symbol, period_cur)
            _, _, macdhist = macd_indicator.clean_macd(kline.close)
            if not macd_indicator.is_expand_golden_cross(macdhist):
                continue

        # Check grandparent period.
        time.sleep(args.interval)
        period_cur = constants.PERIODS[idx + 2]
        kline = XueQiu.kline(symbol, period_cur)
        _, _, macdhist = macd_indicator.clean_macd(kline.close)
        if not macd_indicator.is_expand_golden_cross(macdhist):
            continue

        # Check current period.
        time.sleep(args.interval)
        kline = XueQiu.kline(symbol, args.period)
        macd, macdsignal, macdhist = macd_indicator.clean_macd(kline.close)
        if not macd_indicator.is_negative(macd, macdsignal, args.strict):
            continue
        if not macd_indicator.is_about_to_golden_cross(macdhist, args.strict):
            continue

        # Check child period.
        if args.strict:
            time.sleep(args.interval)
            period_cur = constants.PERIODS[idx - 1]
            kline = XueQiu.kline(symbol, period_cur)
            macd, macdsignal, macdhist = macd_indicator.clean_macd(kline.close)
            if not macd_indicator.is_negative(macd, macdsignal, args.strict):
                continue
            if not macd_indicator.is_about_to_bottom_divergence(
                    macd, macdsignal, macd_indicator, args.strict):
                continue

        res.add(symbol)
    return stocks[stocks.index.isin(res)].sort_index()


def main():
    """Indicate buy signals."""
    args = parse_args()

    if args.period not in constants.PERIODS_VALID:
        raise exceptions.InvalidPeriodError()

    stocks = load_stocks(args.ebk)
    stocks = check_buy(stocks, args)

    utils.export_ebk(stocks.index, args.output)

    title = args.title
    if not title:
        title = f'{args.period}级别可能买点'
        if stocks.index.empty:
            title = f'{args.period}级别无可能买点'

    content = [title]
    content.extend([
        f'{symbol} {stocks.loc[symbol, "name"]}'
        for symbol in stocks.index
    ])
    content.append(str(datetime.utcnow() + timedelta(hours=8)))
    content = '\n'.join(content)
    print(content)
    utils.send_notification(content, args.token, args.topic_ids, args.uids)


if __name__ == '__main__':
    main()
